@@ -1,5 +1,6 @@
 /* objcopy.c -- copy object file from input to output, optionally massaging it.
-   Copyright (C) 1991, 92, 93, 94, 95, 1996 Free Software Foundation, Inc.
+   Copyright (C) 1991, 92, 93, 94, 95, 96, 97, 1998
+   Free Software Foundation, Inc.
 
    This file is part of GNU Binutils.
 
@@ -15,7 +16,8 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
+   02111-1307, USA.  */
 
 #include "bfd.h"
 #include "progress.h"
@@ -25,19 +27,62 @@
 #include "budbg.h"
 #include <sys/stat.h>
 
+#ifdef HAVE_GOOD_UTIME_H
+#include <utime.h>
+#else /* ! HAVE_GOOD_UTIME_H */
+#ifdef HAVE_UTIMES
+#include <sys/time.h>
+#endif /* HAVE_UTIMES */
+#endif /* ! HAVE_GOOD_UTIME_H */
+
+/* A list of symbols to explicitly strip out, or to keep.  A linked
+   list is good enough for a small number from the command line, but
+   this will slow things down a lot if many symbols are being
+   deleted. */
+
+struct symlist
+{
+  const char *name;
+  struct symlist *next;
+};
+
+/* List invented to support redefine_sym */
+
+struct redefine_list_node
+{
+  char *source;
+  char *target;
+  struct redefine_list_node *next;
+};
+
+static void copy_usage PARAMS ((FILE *, int));
+static void strip_usage PARAMS ((FILE *, int));
 static flagword parse_flags PARAMS ((const char *));
 static struct section_list *find_section_list PARAMS ((const char *, boolean));
 static void setup_section PARAMS ((bfd *, asection *, PTR));
 static void copy_section PARAMS ((bfd *, asection *, PTR));
 static void get_sections PARAMS ((bfd *, asection *, PTR));
-static int compare_section_vma PARAMS ((const PTR, const PTR));
-static void add_strip_symbol PARAMS ((const char *));
-static boolean is_strip_symbol PARAMS ((const char *));
+static int compare_section_lma PARAMS ((const PTR, const PTR));
+static void add_specific_symbol PARAMS ((const char *, struct symlist **));
+static boolean is_specified_symbol PARAMS ((const char *, struct symlist *));
 static boolean is_strip_section PARAMS ((bfd *, asection *));
 static unsigned int filter_symbols
   PARAMS ((bfd *, bfd *, asymbol **, asymbol **, long));
 static void mark_symbols_used_in_relocations PARAMS ((bfd *, asection *, PTR));
+static void filter_bytes PARAMS ((char *, bfd_size_type *));
 static boolean write_debugging_info PARAMS ((bfd *, PTR, long *, asymbol ***));
+static void copy_object PARAMS ((bfd *, bfd *));
+static void copy_archive PARAMS ((bfd *, bfd *, const char *));
+static void copy_file
+  PARAMS ((const char *, const char *, const char *, const char *));
+static int simple_copy PARAMS ((const char *, const char *));
+static int smart_rename PARAMS ((const char *, const char *));
+static void set_times PARAMS ((const char *, const struct stat *));
+static int strip_main PARAMS ((int, char **));
+static int copy_main PARAMS ((int, char **));
+
+static char* lookup_sym_redefinition PARAMS((char *source));
+static void redefine_list_append PARAMS ((const char *, const char *));
 
 #define nonfatal(s) {bfd_nonfatal(s); status = 1; return;}
 
@@ -146,6 +191,18 @@ static boolean change_leading_char = false;
 
 static boolean remove_leading_char = false;
 
+/* List of symbols to strip, keep, localize, and weaken.  */
+
+static struct symlist *strip_specific_list = NULL;
+static struct symlist *keep_specific_list = NULL;
+static struct symlist *localize_specific_list = NULL;
+static struct symlist *weaken_specific_list = NULL;
+static struct redefine_list_node *redefine_sym_list = NULL;
+
+/* If this is true, we weaken global symbols (set BSF_WEAK).  */
+
+static boolean weaken = false;
+
 /* 150 isn't special; it's just an arbitrary non-ASCII char value.  */
 
 #define OPTION_ADD_SECTION 150
@@ -163,6 +220,7 @@ static boolean remove_leading_char = false;
 #define OPTION_SET_START (OPTION_SET_SECTION_FLAGS + 1)
 #define OPTION_STRIP_UNNEEDED (OPTION_SET_START + 1)
 #define OPTION_WEAKEN (OPTION_STRIP_UNNEEDED + 1)
+#define OPTION_REDEFINE_SYM (OPTION_WEAKEN + 1) 
 
 /* Options to handle if running as "strip".  */
 
@@ -177,6 +235,7 @@ static struct option strip_options[] =
   {"keep-symbol", required_argument, 0, 'K'},
   {"output-format", required_argument, 0, 'O'},	/* Obsolete */
   {"output-target", required_argument, 0, 'O'},
+  {"preserve-dates", no_argument, 0, 'p'},
   {"remove-section", required_argument, 0, 'R'},
   {"strip-all", no_argument, 0, 's'},
   {"strip-debug", no_argument, 0, 'S'},
@@ -213,6 +272,8 @@ static struct option copy_options[] =
   {"output-format", required_argument, 0, 'O'},	/* Obsolete */
   {"output-target", required_argument, 0, 'O'},
   {"pad-to", required_argument, 0, OPTION_PAD_TO},
+  {"preserve-dates", no_argument, 0, 'p'},
+  {"localize-symbol", required_argument, 0, 'L'},
   {"remove-leading-char", no_argument, 0, OPTION_REMOVE_LEADING_CHAR},
   {"remove-section", required_argument, 0, 'R'},
   {"set-section-flags", required_argument, 0, OPTION_SET_SECTION_FLAGS},
@@ -225,6 +286,8 @@ static struct option copy_options[] =
   {"verbose", no_argument, 0, 'v'},
   {"version", no_argument, 0, 'V'},
   {"weaken", no_argument, 0, OPTION_WEAKEN},
+  {"weaken-symbol", required_argument, 0, 'W'},
+  {"redefine-sym", required_argument, 0, OPTION_REDEFINE_SYM},
   {0, no_argument, 0, 0}
 };
 
@@ -236,6 +299,7 @@ extern char *program_name;
    -1 means if we should use argv[0] to decide. */
 extern int is_strip;
 
+extern void binutils_bug_report PARAMS((FILE *));
 
 static void
 copy_usage (stream, exit_status)
@@ -243,24 +307,26 @@ copy_usage (stream, exit_status)
      int exit_status;
 {
   fprintf (stream, "\
-Usage: %s [-vVSgxX] [-I bfdname] [-O bfdname] [-F bfdname] [-b byte]\n\
+Usage: %s [-vVSpgxX] [-I bfdname] [-O bfdname] [-F bfdname] [-b byte]\n\
        [-R section] [-i interleave] [--interleave=interleave] [--byte=byte]\n\
        [--input-target=bfdname] [--output-target=bfdname] [--target=bfdname]\n\
        [--strip-all] [--strip-debug] [--strip-unneeded] [--discard-all]\n\
        [--discard-locals] [--debugging] [--remove-section=section]\n",
 	   program_name);
   fprintf (stream, "\
-       [--gap-fill=val] [--pad-to=address]\n\
+       [--gap-fill=val] [--pad-to=address] [--preserve-dates]\n\
        [--set-start=val] [--adjust-start=incr]\n\
        [--adjust-vma=incr] [--adjust-section-vma=section{=,+,-}val]\n\
        [--adjust-warnings] [--no-adjust-warnings]\n\
        [--set-section-flags=section=flags] [--add-section=sectionname=filename]\n\
        [--keep-symbol symbol] [-K symbol] [--strip-symbol symbol] [-N symbol]\n\
-       [--change-leading-char] [--remove-leading-char] [--weaken] [--verbose]\n\
-       [--version] [--help] in-file [out-file]\n");
+       [--localize-symbol symbol] [-L symbol] [--weaken-symbol symbol]\n\
+       [--redefine-sym=old_form=new_form] [-W symbol]\n\
+       [--change-leading-char] [--remove-leading-char] [--weaken]\n\
+       [--verbose] [--version] [--help] in-file [out-file]\n");
   list_supported_targets (program_name, stream);
   if (exit_status == 0)
-    fprintf (stream, "Report bugs to bug-gnu-utils@prep.ai.mit.edu\n");
+    binutils_bug_report (stream);
   exit (exit_status);
 }
 
@@ -270,16 +336,16 @@ strip_usage (stream, exit_status)
      int exit_status;
 {
   fprintf (stream, "\
-Usage: %s [-vVsSgxX] [-I bfdname] [-O bfdname] [-F bfdname] [-R section]\n\
+Usage: %s [-vVsSpgxX] [-I bfdname] [-O bfdname] [-F bfdname] [-R section]\n\
        [--input-target=bfdname] [--output-target=bfdname] [--target=bfdname]\n\
        [--strip-all] [--strip-debug] [--strip-unneeded] [--discard-all]\n\
        [--discard-locals] [--keep-symbol symbol] [-K symbol]\n\
        [--strip-symbol symbol] [-N symbol] [--remove-section=section]\n\
-       [-o file] [--verbose] [--version] [--help] file...\n",
+       [-o file] [--preserve-dates] [--verbose] [--version] [--help] file...\n",
 	   program_name);
   list_supported_targets (program_name, stream);
   if (exit_status == 0)
-    fprintf (stream, "Report bugs to bug-gnu-utils@prep.ai.mit.edu\n");
+    binutils_bug_report (stream);
   exit (exit_status);
 }
 
@@ -307,14 +373,31 @@ parse_flags (s)
 	  ++snext;
 	}
 
-#define PARSE_FLAG(fname,fval) if (strncmp (fname, s, len) == 0) ret |= fval;
+      if (0) ;
+#define PARSE_FLAG(fname,fval) \
+  else if (strncasecmp (fname, s, len) == 0) ret |= fval
       PARSE_FLAG ("alloc", SEC_ALLOC);
       PARSE_FLAG ("load", SEC_LOAD);
       PARSE_FLAG ("readonly", SEC_READONLY);
       PARSE_FLAG ("code", SEC_CODE);
       PARSE_FLAG ("data", SEC_DATA);
       PARSE_FLAG ("rom", SEC_ROM);
+      PARSE_FLAG ("contents", SEC_HAS_CONTENTS);
 #undef PARSE_FLAG
+      else
+	{
+	  char *copy;
+
+	  copy = xmalloc (len + 1);
+	  strncpy (copy, s, len);
+	  copy[len] = '\0';
+	  fprintf (stderr, "%s: unrecognized section flag `%s'\n",
+		   program_name, copy);
+	  fprintf (stderr,
+		   "%s: supported flags: alloc, load, readonly, code, data, rom, contents\n",
+		   program_name);
+	  exit (1);
+	}
 
       s = snext;
     }
@@ -354,59 +437,37 @@ find_section_list (name, add)
   return p;
 }
 
-/* Make a list of symbols to explicitly strip out, or to keep.  A
-   linked list is good enough for a small number from the command
-   line, but this will slow things down a lot if many symbols are
-   being deleted. */
-
-struct symlist
-{
-  const char *name;
-  struct symlist *next;
-};
-
-/* List of symbols to strip.  */
-
-static struct symlist *strip_specific_list = NULL;
-
-/* If this is false, we strip the symbols in strip_specific_list.
-   Otherwise, we keep only the symbols in the list.  */
-
-static boolean keep_symbols = false;
-
-/* If this is true, we weaken global symbols (set BSF_WEAK).  */
-
-static boolean weaken = false;
-
 /* Add a symbol to strip_specific_list.  */
 
 static void 
-add_strip_symbol (name)
+add_specific_symbol (name, list)
      const char *name;
+     struct symlist **list;
 {
   struct symlist *tmp_list;
 
   tmp_list = (struct symlist *) xmalloc (sizeof (struct symlist));
   tmp_list->name = name;
-  tmp_list->next = strip_specific_list;
-  strip_specific_list = tmp_list;
+  tmp_list->next = *list;
+  *list = tmp_list;
 }
 
 /* See whether a symbol should be stripped or kept based on
    strip_specific_list and keep_symbols.  */
 
 static boolean
-is_strip_symbol (name)
+is_specified_symbol (name, list)
      const char *name;
+     struct symlist *list;
 {
   struct symlist *tmp_list;
 
-  for (tmp_list = strip_specific_list; tmp_list; tmp_list = tmp_list->next)
+  for (tmp_list = list; tmp_list; tmp_list = tmp_list->next)
     {
       if (strcmp (name, tmp_list->name) == 0)
-	return keep_symbols ? false : true;
+	return true;
     }
-  return keep_symbols;
+  return false;
 }
 
 /* See if a section is being removed.  */
@@ -434,7 +495,8 @@ is_strip_section (abfd, sec)
 
 /* Choose which symbol entries to copy; put the result in OSYMS.
    We don't copy in place, because that confuses the relocs.
-   Return the number of symbols to print.  */
+   Return the number of symbols to print.  The redefine_sym option 
+   is just another filter, not unlike remove_leading_char */
 
 static unsigned int
 filter_symbols (abfd, obfd, osyms, isyms, symcount)
@@ -450,28 +512,37 @@ filter_symbols (abfd, obfd, osyms, isyms, symcount)
     {
       asymbol *sym = from[src_count];
       flagword flags = sym->flags;
+      const char *name = bfd_asymbol_name (sym);
       int keep;
 
+	if(redefine_sym_list)
+	{
+		char *old_name, *new_name;
+		
+		old_name =  (char*) bfd_asymbol_name (sym);
+		new_name = lookup_sym_redefinition(old_name);
+		name = bfd_asymbol_name (sym) = new_name;
+	}
+
       if (change_leading_char
 	  && (bfd_get_symbol_leading_char (abfd)
 	      != bfd_get_symbol_leading_char (obfd))
 	  && (bfd_get_symbol_leading_char (abfd) == '\0'
-	      || (bfd_asymbol_name (sym)[0]
-		  == bfd_get_symbol_leading_char (abfd))))
+	      || (name[0] == bfd_get_symbol_leading_char (abfd))))
 	{
 	  if (bfd_get_symbol_leading_char (obfd) == '\0')
-	    bfd_asymbol_name (sym) = bfd_asymbol_name (sym) + 1;
+	    name = bfd_asymbol_name (sym) = name + 1;
 	  else
 	    {
 	      char *n;
 
-	      n = xmalloc (strlen (bfd_asymbol_name (sym)) + 2);
+	      n = xmalloc (strlen (name) + 2);
 	      n[0] = bfd_get_symbol_leading_char (obfd);
 	      if (bfd_get_symbol_leading_char (abfd) == '\0')
-		strcpy (n + 1, bfd_asymbol_name (sym));
+		strcpy (n + 1, name);
 	      else
-		strcpy (n + 1, bfd_asymbol_name (sym) + 1);
-	      bfd_asymbol_name (sym) = n;
+		strcpy (n + 1, name + 1);
+	      name = bfd_asymbol_name (sym) = n;
 	    }
 	}
 
@@ -480,8 +551,8 @@ filter_symbols (abfd, obfd, osyms, isyms, symcount)
 	      || (flags & BSF_WEAK) != 0
 	      || bfd_is_und_section (bfd_get_section (sym))
 	      || bfd_is_com_section (bfd_get_section (sym)))
-	  && bfd_asymbol_name (sym)[0] == bfd_get_symbol_leading_char (abfd))
-	bfd_asymbol_name (sym) = bfd_asymbol_name (sym) + 1;
+	  && name[0] == bfd_get_symbol_leading_char (abfd))
+	name = bfd_asymbol_name (sym) = name + 1;
 
       if ((flags & BSF_KEEP) != 0)		/* Used in relocation.  */
 	keep = 1;
@@ -500,16 +571,25 @@ filter_symbols (abfd, obfd, osyms, isyms, symcount)
 		    && (discard_locals != locals_start_L
 			|| ! bfd_is_local_label (abfd, sym))));
 
-      if (keep && is_strip_symbol (bfd_asymbol_name (sym)))
+      if (keep && is_specified_symbol (name, strip_specific_list))
 	keep = 0;
+      if (!keep && is_specified_symbol (name, keep_specific_list))
+	keep = 1;
       if (keep && is_strip_section (abfd, bfd_get_section (sym)))
 	keep = 0;
 
-      if (keep && weaken && (flags & BSF_GLOBAL) != 0)
+      if (keep && (flags & BSF_GLOBAL) != 0
+	  && (weaken || is_specified_symbol (name, weaken_specific_list)))
 	{
 	  sym->flags &=~ BSF_GLOBAL;
 	  sym->flags |= BSF_WEAK;
 	}
+      if (keep && (flags & (BSF_GLOBAL | BSF_WEAK))
+	  && is_specified_symbol (name, localize_specific_list))
+	{
+	  sym->flags &= ~(BSF_GLOBAL | BSF_WEAK);
+	  sym->flags |= BSF_LOCAL;
+	}
 
       if (keep)
 	to[dst_count++] = sym;
@@ -520,10 +600,88 @@ filter_symbols (abfd, obfd, osyms, isyms, symcount)
   return dst_count;
 }
 
+static char*
+lookup_sym_redefinition(source)
+	char *source;
+{
+	char* result;
+	struct redefine_list_node *list;
+	
+	if(!redefine_sym_list) return source;
+
+	result = source;
+	list = redefine_sym_list;
+	
+	while(list != NULL)
+	{
+		if(strcmp(source, list->source) == 0)
+		{
+			result = list->target;
+			break;
+		}
+		
+		list = list->next;
+	}
+		
+	return result;
+}
+
+/* Add a node to a symbol redefine list */
+
+static void
+redefine_list_append(source, target)
+	const char *source;
+	const char *target;
+{
+
+	struct redefine_list_node *list;
+	struct redefine_list_node *new_node;
+    		
+	new_node = (struct redefine_list_node*) xmalloc (sizeof(struct redefine_list_node));
+  
+	new_node->source = strdup(source);
+	new_node->target = strdup(target);
+	new_node->next = NULL; 
+
+	if(!redefine_sym_list)  
+	{   
+		redefine_sym_list = new_node;
+	}
+	else
+	{
+		list = redefine_sym_list;	
+		
+		while(list != NULL)
+		{
+
+			if(strcmp(source, list->source) == 0)
+			{
+				fprintf(stderr, "redefine-sym:  Multiple redefinition of symbol \"%s\"\n", source);
+				exit(1);
+			}
+
+			if(strcmp(target, list->target) == 0)
+			{
+				fprintf(stderr, "redefine-sym: Symbol \"%s\" is target of more than one redefinition", target);
+				exit(1);
+			} 
+
+			if(list->next == NULL)
+			{
+				list->next = new_node;
+				list = list->next;
+			}
+
+			list = list->next;
+		}
+	} 
+}
+
+
 /* Keep only every `copy_byte'th byte in MEMHUNK, which is *SIZE bytes long.
    Adjust *SIZE.  */
 
-void
+static void
 filter_bytes (memhunk, size)
      char *memhunk;
      bfd_size_type *size;
@@ -656,7 +814,7 @@ copy_object (ibfd, obfd)
       set = osections;
       bfd_map_over_sections (obfd, get_sections, (void *) &set);
 
-      qsort (osections, c, sizeof (asection *), compare_section_vma);
+      qsort (osections, c, sizeof (asection *), compare_section_lma);
 
       gaps = (bfd_size_type *) xmalloc (c * sizeof (bfd_size_type));
       memset (gaps, 0, c * sizeof (bfd_size_type));
@@ -675,8 +833,8 @@ copy_object (ibfd, obfd)
 		continue;
 
 	      size = bfd_section_size (obfd, osections[i]);
-	      gap_start = bfd_section_vma (obfd, osections[i]) + size;
-	      gap_stop = bfd_section_vma (obfd, osections[i + 1]);
+	      gap_start = bfd_section_lma (obfd, osections[i]) + size;
+	      gap_stop = bfd_section_lma (obfd, osections[i + 1]);
 	      if (gap_start < gap_stop)
 		{
 		  if (! bfd_set_section_size (obfd, osections[i],
@@ -698,15 +856,15 @@ copy_object (ibfd, obfd)
 
       if (pad_to_set)
 	{
-	  bfd_vma vma;
+	  bfd_vma lma;
 	  bfd_size_type size;
 
-	  vma = bfd_section_vma (obfd, osections[c - 1]);
+	  lma = bfd_section_lma (obfd, osections[c - 1]);
 	  size = bfd_section_size (obfd, osections[c - 1]);
-	  if (vma + size < pad_to)
+	  if (lma + size < pad_to)
 	    {
 	      if (! bfd_set_section_size (obfd, osections[c - 1],
-					  pad_to - vma))
+					  pad_to - lma))
 		{
 		  fprintf (stderr, "%s: Can't add padding to %s: %s\n",
 			   program_name,
@@ -716,12 +874,12 @@ copy_object (ibfd, obfd)
 		}
 	      else
 		{
-		  gaps[c - 1] = pad_to - (vma + size);
-		  if (max_gap < pad_to - (vma + size))
-		    max_gap = pad_to - (vma + size);
+		  gaps[c - 1] = pad_to - (lma + size);
+		  if (max_gap < pad_to - (lma + size))
+		    max_gap = pad_to - (lma + size);
 		}
 	    }
-	}	      
+	}
     }
 
   /* Symbol filtering must happen after the output sections have
@@ -756,10 +914,14 @@ copy_object (ibfd, obfd)
 	  || strip_symbols == strip_unneeded
 	  || discard_locals != locals_undef
 	  || strip_specific_list != NULL
+	  || keep_specific_list != NULL
+	  || localize_specific_list != NULL
+	  || weaken_specific_list != NULL
 	  || sections_removed
 	  || convert_debugging
 	  || change_leading_char
 	  || remove_leading_char
+	  || redefine_sym_list  
 	  || weaken)
 	{
 	  /* Mark symbols used in output relocations so that they
@@ -862,21 +1024,6 @@ copy_object (ibfd, obfd)
     }
 }
 
-static char *
-cat (a, b, c)
-     char *a;
-     char *b;
-     char *c;
-{
-  size_t size = strlen (a) + strlen (b) + strlen (c);
-  char *r = xmalloc (size + 1);
-
-  strcpy (r, a);
-  strcat (r, b);
-  strcat (r, c);
-  return r;
-}
-
 /* Read each archive element in turn from IBFD, copy the
    contents to temp file, and keep the temp file handle.  */
 
@@ -884,7 +1031,7 @@ static void
 copy_archive (ibfd, obfd, output_target)
      bfd *ibfd;
      bfd *obfd;
-     char *output_target;
+     const char *output_target;
 {
   struct name_list
     {
@@ -897,7 +1044,11 @@ copy_archive (ibfd, obfd, output_target)
   char *dir = make_tempname (bfd_get_filename (obfd));
 
   /* Make a temp directory to hold the contents.  */
+#if defined (_WIN32) && !defined (__CYGWIN32__)
+  if (mkdir (dir) != 0)
+#else
   if (mkdir (dir, 0700) != 0)
+#endif
     {
       fatal ("cannot mkdir %s for archive copying (error: %s)",
 	     dir, strerror (errno));
@@ -910,7 +1061,8 @@ copy_archive (ibfd, obfd, output_target)
   while (this_element != (bfd *) NULL)
     {
       /* Create an output file for this member.  */
-      char *output_name = cat (dir, "/", bfd_get_filename(this_element));
+      char *output_name = concat (dir, "/", bfd_get_filename(this_element),
+				  (char *) NULL);
       bfd *output_bfd = bfd_openw (output_name, output_target);
       bfd *last_element;
 
@@ -974,10 +1126,10 @@ copy_archive (ibfd, obfd, output_target)
 
 static void
 copy_file (input_filename, output_filename, input_target, output_target)
-     char *input_filename;
-     char *output_filename;
-     char *input_target;
-     char *output_target;
+     const char *input_filename;
+     const char *output_filename;
+     const char *input_target;
+     const char *output_target;
 {
   bfd *ibfd;
   char **matching;
@@ -1249,6 +1401,22 @@ copy_section (ibfd, isection, obfdarg)
 	}
       free (memhunk);
     }
+  else if (p != NULL && p->set_flags && (p->flags & SEC_HAS_CONTENTS) != 0)
+    {
+      PTR memhunk = (PTR) xmalloc ((unsigned) size);
+
+      /* We don't permit the user to turn off the SEC_HAS_CONTENTS
+	 flag--they can just remove the section entirely and add it
+	 back again.  However, we do permit them to turn on the
+	 SEC_HAS_CONTENTS flag, and take it to mean that the section
+	 contents should be zeroed out.  */
+
+      memset (memhunk, 0, size);
+      if (! bfd_set_section_contents (obfd, osection, memhunk, (file_ptr) 0,
+				      size))
+	nonfatal (bfd_get_filename (obfd));
+      free (memhunk);
+    }
 }
 
 /* Get all the sections.  This is used when --gap-fill or --pad-to is
@@ -1271,7 +1439,7 @@ get_sections (obfd, osection, secppparg)
    sections to the front, where they are easier to ignore.  */
 
 static int
-compare_section_vma (arg1, arg2)
+compare_section_lma (arg1, arg2)
      const PTR arg1;
      const PTR arg2;
 {
@@ -1296,13 +1464,13 @@ compare_section_vma (arg1, arg2)
 	return 1;
     }
 
-  /* Sort sections by VMA.  */
-  if ((*sec1)->vma > (*sec2)->vma)
+  /* Sort sections by LMA.  */
+  if ((*sec1)->lma > (*sec2)->lma)
     return 1;
-  else if ((*sec1)->vma < (*sec2)->vma)
+  else if ((*sec1)->lma < (*sec2)->lma)
     return -1;
 
-  /* Sort sections with the same VMA by size.  */
+  /* Sort sections with the same LMA by size.  */
   if ((*sec1)->_raw_size > (*sec2)->_raw_size)
     return 1;
   else if ((*sec1)->_raw_size < (*sec2)->_raw_size)
@@ -1434,7 +1602,8 @@ write_debugging_info (obfd, dhandle, symcountp, symppp)
 
 static int
 simple_copy (from, to)
-     char *from, *to;
+     const char *from;
+     const char *to;
 {
   int fromfd, tofd, nread;
   int saved;
@@ -1488,7 +1657,8 @@ simple_copy (from, to)
 
 static int
 smart_rename (from, to)
-     char *from, *to;
+     const char *from;
+     const char *to;
 {
   struct stat s;
   int ret = 0;
@@ -1496,6 +1666,24 @@ smart_rename (from, to)
   if (lstat (to, &s))
     return -1;
 
+#if defined (_WIN32) && !defined (__CYGWIN32__)
+  /* Win32, unlike unix, will not erase `to' in `rename(from, to)' but
+     fail instead.  Also, chown is not present.  */
+
+  if (stat (to, &s) == 0)
+    remove (to);
+
+  ret = rename (from, to);
+  if (ret != 0)
+    {
+      /* We have to clean up here. */
+      int saved = errno;
+      fprintf (stderr, "%s: %s: ", program_name, to);
+      errno = saved;
+      perror ("rename");
+      unlink (from);
+    }
+#else
   /* Use rename only if TO is not a symbolic link and has
      only one hard link.  */
   if (!S_ISLNK (s.st_mode) && s.st_nlink == 1)
@@ -1540,9 +1728,54 @@ smart_rename (from, to)
 	}
       unlink (from);
     }
+#endif /* _WIN32 && !__CYGWIN32__ */
+
   return ret;
 }
 
+/* Set the times of the file DESTINATION to be the same as those in
+   STATBUF.  */
+
+static void
+set_times (destination, statbuf)
+     const char *destination;
+     const struct stat *statbuf;
+{
+  int result;
+
+  {
+#ifdef HAVE_GOOD_UTIME_H
+    struct utimbuf tb;
+
+    tb.actime = statbuf->st_atime;
+    tb.modtime = statbuf->st_mtime;
+    result = utime (destination, &tb);
+#else /* ! HAVE_GOOD_UTIME_H */
+#ifndef HAVE_UTIMES
+    long tb[2];
+
+    tb[0] = statbuf->st_atime;
+    tb[1] = statbuf->st_mtime;
+    result = utime (destination, tb);
+#else /* HAVE_UTIMES */
+    struct timeval tv[2];
+
+    tv[0].tv_sec = statbuf->st_atime;
+    tv[0].tv_usec = 0;
+    tv[1].tv_sec = statbuf->st_mtime;
+    tv[1].tv_usec = 0;
+    result = utimes (destination, tv);
+#endif /* HAVE_UTIMES */
+#endif /* ! HAVE_GOOD_UTIME_H */
+  }
+
+  if (result != 0)
+    {
+      fprintf (stderr, "%s: ", destination);
+      perror ("can not set time");
+    }
+}
+
 static int
 strip_main (argc, argv)
      int argc;
@@ -1550,11 +1783,12 @@ strip_main (argc, argv)
 {
   char *input_target = NULL, *output_target = NULL;
   boolean show_version = false;
+  boolean preserve_dates = false;
   int c, i;
   struct section_list *p;
   char *output_file = NULL;
 
-  while ((c = getopt_long (argc, argv, "I:O:F:K:N:R:o:sSgxXVv",
+  while ((c = getopt_long (argc, argv, "I:O:F:K:N:R:o:sSpgxXVv",
 			   strip_options, (int *) 0)) != EOF)
     {
       switch (c)
@@ -1584,27 +1818,17 @@ strip_main (argc, argv)
 	  strip_symbols = strip_unneeded;
 	  break;
 	case 'K':
-	  if (! keep_symbols && strip_specific_list != NULL)
-	    {
-	      fprintf (stderr, "%s: Can not specify both -K and -N\n",
-		       program_name);
-	      strip_usage (stderr, 1);
-	    }
-	  keep_symbols = true;
-	  add_strip_symbol (optarg);
+	  add_specific_symbol (optarg, &keep_specific_list);
 	  break;
 	case 'N':
-	  if (keep_symbols)
-	    {
-	      fprintf (stderr, "%s: Can not specify both -K and -N\n",
-		       program_name);
-	      strip_usage (stderr, 1);
-	    }
-	  add_strip_symbol (optarg);
+	  add_specific_symbol (optarg, &strip_specific_list);
 	  break;
 	case 'o':
 	  output_file = optarg;
 	  break;
+	case 'p':
+	  preserve_dates = true;
+	  break;
 	case 'x':
 	  discard_locals = locals_all;
 	  break;
@@ -1646,16 +1870,30 @@ strip_main (argc, argv)
   for (; i < argc; i++)
     {
       int hold_status = status;
+      struct stat statbuf;
       char *tmpname;
 
+      if (preserve_dates)
+	{
+	  if (stat (argv[i], &statbuf) < 0)
+	    {
+	      fprintf (stderr, "%s: ", argv[i]);
+	      perror ("cannot stat");
+	      continue;
+	    }
+	}
+
       if (output_file != NULL)
 	tmpname = output_file;
       else
 	tmpname = make_tempname (argv[i]);
       status = 0;
+
       copy_file (argv[i], tmpname, input_target, output_target);
       if (status == 0)
 	{
+	  if (preserve_dates)
+	    set_times (tmpname, &statbuf);
 	  if (output_file == NULL)
 	    smart_rename (tmpname, argv[i]);
 	  status = hold_status;
@@ -1678,10 +1916,12 @@ copy_main (argc, argv)
   char *input_target = NULL, *output_target = NULL;
   boolean show_version = false;
   boolean adjust_warn = true;
+  boolean preserve_dates = false;
   int c;
   struct section_list *p;
+  struct stat statbuf;
 
-  while ((c = getopt_long (argc, argv, "b:i:I:K:N:s:O:d:F:R:SgxXVv",
+  while ((c = getopt_long (argc, argv, "b:i:I:K:N:s:O:d:F:L:R:SpgxXVvW:",
 			   copy_options, (int *) 0)) != EOF)
     {
       switch (c)
@@ -1730,23 +1970,19 @@ copy_main (argc, argv)
 	  strip_symbols = strip_unneeded;
 	  break;
 	case 'K':
-	  if (! keep_symbols && strip_specific_list != NULL)
-	    {
-	      fprintf (stderr, "%s: Can not specify both -K and -N\n",
-		       program_name);
-	      strip_usage (stderr, 1);
-	    }
-	  keep_symbols = true;
-	  add_strip_symbol (optarg);
+	  add_specific_symbol (optarg, &keep_specific_list);
 	  break;
 	case 'N':
-	  if (keep_symbols)
-	    {
-	      fprintf (stderr, "%s: Can not specify both -K and -N\n",
-		       program_name);
-	      strip_usage (stderr, 1);
-	    }
-	  add_strip_symbol (optarg);
+	  add_specific_symbol (optarg, &strip_specific_list);
+	  break;
+	case 'L':
+	  add_specific_symbol (optarg, &localize_specific_list);
+	  break;
+	case 'W':
+	  add_specific_symbol (optarg, &weaken_specific_list);
+	  break;
+	case 'p':
+	  preserve_dates = true;
 	  break;
 	case 'x':
 	  discard_locals = locals_all;
@@ -1905,6 +2141,41 @@ copy_main (argc, argv)
 	case OPTION_REMOVE_LEADING_CHAR:
 	  remove_leading_char = true;
 	  break;
+	case OPTION_REDEFINE_SYM:
+	{
+	  /* Push this redefinition onto redefine_symbol_list */
+	  
+	  int len;
+	  char *s;
+	  char* nextarg;
+	  char *source, *target;
+
+	  s = strchr (optarg, '=');
+	    if (s == NULL)
+	      {
+		fprintf (stderr, "%s: bad format for --redefine-sym\n",
+			 program_name);
+		exit (1);
+	      }	 
+	      
+	  len = s - optarg; 
+	  source = (char *) xmalloc (len + 1);
+	  strncpy (source, optarg, len);
+	  source[len] = '\0';
+
+	  nextarg = s + 1;	 
+	  len = strlen(nextarg);
+	  target = (char *) xmalloc (len + 1);
+	  strncpy (target, nextarg, len);
+	  target[len] = '\0';
+
+	  redefine_list_append(source, target);
+	  
+	  free(source);
+	  free(target);	      
+	      
+	 } 	
+	 break;
 	case OPTION_SET_SECTION_FLAGS:
 	  {
 	    const char *s;
@@ -1967,21 +2238,38 @@ copy_main (argc, argv)
   if (output_target == (char *) NULL)
     output_target = input_target;
 
+  if (preserve_dates)
+    {
+      if (stat (input_filename, &statbuf) < 0)
+	{
+	  fprintf (stderr, "%s: ", input_filename);
+	  perror ("cannot stat");
+	  exit (1);
+	}
+    }
+
   /* If there is no destination file then create a temp and rename
      the result into the input.  */
 
   if (output_filename == (char *) NULL)
     {
       char *tmpname = make_tempname (input_filename);
+
       copy_file (input_filename, tmpname, input_target, output_target);
       if (status == 0)
-	smart_rename (tmpname, input_filename);
+	{	
+	  if (preserve_dates)
+	    set_times (tmpname, &statbuf);
+	  smart_rename (tmpname, input_filename);
+	}
       else
 	unlink (tmpname);
     }
   else
     {
       copy_file (input_filename, output_filename, input_target, output_target);
+      if (status == 0 && preserve_dates)
+	set_times (output_filename, &statbuf);
     }
 
   if (adjust_warn)
@@ -2016,6 +2304,7 @@ main (argc, argv)
   discard_locals = locals_undef;
 
   bfd_init ();
+  set_default_bfd_target ();
 
   if (is_strip < 0)
     {
