@@ -73,6 +73,9 @@ static bfd_vma stop_address = (bfd_vma) -1;  /* --stop-address */
 static int dump_debugging;		/* --debugging */
 static bfd_vma adjust_section_vma = 0;	/* --adjust-vma */
 
+/* --lang=xxx */
+static int demangle_lang = DMGL_ANSI | DMGL_PARAMS;
+
 /* Extra info to pass to the disassembler address printing function.  */
 struct objdump_disasm_info {
   bfd *abfd;
@@ -178,7 +181,7 @@ show_line PARAMS ((bfd *, asection *, bfd_vma));
 
 static void
 disassemble_bytes PARAMS ((struct disassemble_info *, disassembler_ftype,
-			   boolean, bfd_byte *, long, long, arelent ***,
+			   boolean, bfd_byte *, unsigned long, unsigned long, arelent ***,
 			   arelent **));
 
 static void
@@ -211,6 +214,8 @@ read_section_stabs PARAMS ((bfd *, const char *, const char *));
 static void
 print_section_stabs PARAMS ((bfd *, const char *, const char *));
 
+extern void binutils_bug_report PARAMS((FILE *));
+
 static void
 usage (stream, status)
      FILE *stream;
@@ -228,12 +233,12 @@ Usage: %s [-ahifCdDprRtTxsSlw] [-b bfdname] [-m machine] [-j section-name]\n\
        [--syms] [--all-headers] [--dynamic-syms] [--dynamic-reloc]\n\
        [--wide] [--version] [--help] [--private-headers]\n\
        [--start-address=addr] [--stop-address=addr]\n\
-       [--prefix-addresses] [--[no-]show-raw-insn] [--demangle]\n\
+       [--prefix-addresses] [--[no-]show-raw-insn] [--demangle] [--lang=language]\n\
        [--adjust-vma=offset] [-EB|-EL] [--endian={big|little}] objfile...\n\
 at least one option besides -l (--line-numbers) must be given\n");
   list_supported_targets (program_name, stream);
   if (status == 0)
-    fprintf (stream, "Report bugs to bug-gnu-utils@gnu.org\n");
+    binutils_bug_report (stream);
   exit (status);
 }
 
@@ -243,6 +248,7 @@ at least one option besides -l (--line-numbers) must be given\n");
 #define OPTION_START_ADDRESS (OPTION_ENDIAN + 1)
 #define OPTION_STOP_ADDRESS (OPTION_START_ADDRESS + 1)
 #define OPTION_ADJUST_VMA (OPTION_STOP_ADDRESS + 1)
+#define OPTION_LANG (OPTION_ADJUST_VMA + 1)
 
 static struct option long_options[]=
 {
@@ -279,6 +285,7 @@ static struct option long_options[]=
   {"target", required_argument, NULL, 'b'},
   {"version", no_argument, &show_version, 1},
   {"wide", no_argument, &wide_output, 'w'},
+  {"lang", required_argument, NULL, OPTION_LANG},
   {0, no_argument, 0, 0}
 };
 
@@ -641,7 +648,7 @@ objdump_print_symname (abfd, info, sym)
       if (bfd_get_symbol_leading_char (abfd) == name[0])
 	++name;
 
-      alloc = cplus_demangle (name, DMGL_ANSI | DMGL_PARAMS);
+      alloc = cplus_demangle (name, demangle_lang);
       if (alloc == NULL)
 	print = name;
       else
@@ -1153,8 +1160,8 @@ disassemble_bytes (info, disassemble_fn, insns, data, start, stop, relppp,
      disassembler_ftype disassemble_fn;
      boolean insns;
      bfd_byte *data;
-     long start;
-     long stop;
+     unsigned long start;
+     unsigned long stop;
      arelent ***relppp;
      arelent **relppend;
 {
@@ -1163,7 +1170,7 @@ disassemble_bytes (info, disassemble_fn, insns, data, start, stop, relppp,
   int bytes_per_line;
   boolean done_dot;
   int skip_addr_chars;
-  long i;
+  unsigned long i;
 
   aux = (struct objdump_disasm_info *) info->application_data;
   section = aux->sec;
@@ -1200,7 +1207,7 @@ disassemble_bytes (info, disassemble_fn, insns, data, start, stop, relppp,
   i = start;
   while (i < stop)
     {
-      long z;
+      unsigned long z;
       int bytes;
       boolean need_nl = false;
 
@@ -2218,7 +2225,7 @@ dump_symbols (abfd, dynamic)
 		  n = name;
 		  if (bfd_get_symbol_leading_char (cur_bfd) == *n)
 		    ++n;
-		  alloc = cplus_demangle (n, DMGL_ANSI | DMGL_PARAMS);
+		  alloc = cplus_demangle (n, demangle_lang);
 		  if (alloc != NULL)
 		    (*current)->name = alloc;
 		  else
@@ -2748,6 +2755,10 @@ main (argc, argv)
 	      usage (stderr, 1);
 	    }
 	  break;
+	case OPTION_LANG:
+	  if (strcasecmp (optarg, "gnat") == 0)
+	    demangle_lang = DMGL_GNAT;
+	  break;
 	default:
 	  usage (stderr, 1);
 	}
