@@ -129,7 +129,9 @@ fetch_data (info, addr)
 
 #define Sw OP_SEG, w_mode
 #define Ap OP_DIR, lptr
+#if 0
 #define Av OP_DIR, v_mode
+#endif
 #define Ob OP_OFF, b_mode
 #define Ov OP_OFF, v_mode
 #define Xb OP_DSSI, b_mode
@@ -145,8 +147,14 @@ fetch_data (info, addr)
 #define gs OP_REG, gs_reg
 
 #define MX OP_MMX, 0
+#define XM OP_XMM, 0
 #define EM OP_EM, v_mode
+#define EX OP_EX, v_mode
 #define MS OP_MS, b_mode
+#define None OP_None, 0
+#define TD OP_3DNow, 0
+#define SUF OP_3DNowSuffix, 0
+#define PF OP_PREFETCH, b_mode
 
 typedef int (*op_rtn) PARAMS ((int bytemode, int aflag, int dflag));
 
@@ -172,8 +180,14 @@ static int OP_STi  PARAMS ((int, int, int));
 static int OP_ONE PARAMS ((int, int, int));
 #endif
 static int OP_MMX PARAMS ((int, int, int));
+static int OP_XMM PARAMS ((int, int, int));
 static int OP_EM PARAMS ((int, int, int));
+static int OP_EX PARAMS ((int, int, int));
 static int OP_MS PARAMS ((int, int, int));
+static int OP_None PARAMS ((int, int, int));
+static int OP_3DNow PARAMS ((int, int, int));
+static int OP_3DNowSuffix PARAMS ((int, int, int));
+static int OP_PREFETCH PARAMS ((int, int, int));
 
 static void append_prefix PARAMS ((void));
 static void set_op PARAMS ((int op));
@@ -225,26 +239,47 @@ static void ckprefix PARAMS ((void));
 
 #define indir_dx_reg 150
 
-#define GRP1b NULL, NULL, 0
-#define GRP1S NULL, NULL, 1
-#define GRP1Ss NULL, NULL, 2
-#define GRP2b NULL, NULL, 3
-#define GRP2S NULL, NULL, 4
-#define GRP2b_one NULL, NULL, 5
-#define GRP2S_one NULL, NULL, 6
-#define GRP2b_cl NULL, NULL, 7
-#define GRP2S_cl NULL, NULL, 8
-#define GRP3b NULL, NULL, 9
-#define GRP3S NULL, NULL, 10
-#define GRP4  NULL, NULL, 11
-#define GRP5  NULL, NULL, 12
-#define GRP6  NULL, NULL, 13
-#define GRP7 NULL, NULL, 14
-#define GRP8 NULL, NULL, 15
-#define GRP9 NULL, NULL, 16
-#define GRP10 NULL, NULL, 17
-#define GRP11 NULL, NULL, 18
-#define GRP12 NULL, NULL, 19
+#define USE_GROUPS 1
+#define USE_PREFIX_USER_TABLE 2
+
+#define GRP1b NULL, NULL, 0, NULL, USE_GROUPS
+#define GRP1S NULL, NULL, 1, NULL, USE_GROUPS
+#define GRP1Ss NULL, NULL, 2, NULL, USE_GROUPS
+#define GRP2b NULL, NULL, 3, NULL, USE_GROUPS
+#define GRP2S NULL, NULL, 4, NULL, USE_GROUPS
+#define GRP2b_one NULL, NULL, 5, NULL, USE_GROUPS
+#define GRP2S_one NULL, NULL, 6, NULL, USE_GROUPS
+#define GRP2b_cl NULL, NULL, 7, NULL, USE_GROUPS
+#define GRP2S_cl NULL, NULL, 8, NULL, USE_GROUPS
+#define GRP3b NULL, NULL, 9, NULL, USE_GROUPS
+#define GRP3S NULL, NULL, 10, NULL, USE_GROUPS
+#define GRP4  NULL, NULL, 11, NULL, USE_GROUPS
+#define GRP5  NULL, NULL, 12, NULL, USE_GROUPS
+#define GRP6  NULL, NULL, 13, NULL, USE_GROUPS
+#define GRP7 NULL, NULL, 14, NULL, USE_GROUPS
+#define GRP8 NULL, NULL, 15, NULL, USE_GROUPS
+#define GRP9 NULL, NULL, 16, NULL, USE_GROUPS
+#define GRP10 NULL, NULL, 17, NULL, USE_GROUPS
+#define GRP11 NULL, NULL, 18, NULL, USE_GROUPS
+#define GRP12 NULL, NULL, 19, NULL, USE_GROUPS
+#define GRP13 NULL, NULL, 20, NULL, USE_GROUPS
+#define GRP14 NULL, NULL, 21, NULL, USE_GROUPS
+
+#define PREGRP0 NULL, NULL, 0, NULL, USE_PREFIX_USER_TABLE
+#define PREGRP1 NULL, NULL, 1, NULL, USE_PREFIX_USER_TABLE
+#define PREGRP2 NULL, NULL, 2, NULL, USE_PREFIX_USER_TABLE
+#define PREGRP3 NULL, NULL, 3, NULL, USE_PREFIX_USER_TABLE
+#define PREGRP4 NULL, NULL, 4, NULL, USE_PREFIX_USER_TABLE
+#define PREGRP5 NULL, NULL, 5, NULL, USE_PREFIX_USER_TABLE
+#define PREGRP6 NULL, NULL, 6, NULL, USE_PREFIX_USER_TABLE
+#define PREGRP7 NULL, NULL, 7, NULL, USE_PREFIX_USER_TABLE
+#define PREGRP8 NULL, NULL, 8, NULL, USE_PREFIX_USER_TABLE
+#define PREGRP9 NULL, NULL, 9, NULL, USE_PREFIX_USER_TABLE
+#define PREGRP10 NULL, NULL, 10, NULL, USE_PREFIX_USER_TABLE
+#define PREGRP11 NULL, NULL, 11, NULL, USE_PREFIX_USER_TABLE
+#define PREGRP12 NULL, NULL, 12, NULL, USE_PREFIX_USER_TABLE
+#define PREGRP13 NULL, NULL, 13, NULL, USE_PREFIX_USER_TABLE
+#define PREGRP14 NULL, NULL, 14, NULL, USE_PREFIX_USER_TABLE
 
 #define FLOATCODE 50
 #define FLOAT NULL, NULL, FLOATCODE
@@ -369,8 +404,8 @@ static struct dis386 dis386[] = {
   { "popS",	eSI },
   { "popS",	eDI },
   /* 60 */
-  { "pusha" },
-  { "popa" },
+  { "pushaS" },
+  { "popaS" },
   { "boundS",	Gv, Ma },
   { "arpl",	Ew, Gw },
   { "(bad)" },			/* seg fs */
@@ -522,7 +557,7 @@ static struct dis386 dis386[] = {
   { "outb",	Ib, AL },
   { "outS",	Ib, eAX },
   /* e8 */
-  { "call",	Av },
+  { "call",	Jv },
   { "jmp",	Jv },
   { "ljmp",	Ap },
   { "jmp",	Jb },
@@ -564,12 +599,19 @@ static struct dis386 dis386_twobyte[] = {
   { "invd" },
   { "wbinvd" },
   { "(bad)" },  { "ud2a" },  
-  { "(bad)" },  { "(bad)" },  { "(bad)" },  { "(bad)" },  
+  { "(bad)" },  { "prefetch", PF },  { "femms" },  { "3dnow", TD, EM, SUF },  
   /* 10 */
-  { "(bad)" },  { "(bad)" },  { "(bad)" },  { "(bad)" },  
-  { "(bad)" },  { "(bad)" },  { "(bad)" },  { "(bad)" },  
+  { PREGRP8 },
+  { PREGRP9 },
+  { "movlps", XM, Ev },
+  { "movlps", Ev, XM },  
+  { "unpcklps", XM, EX },  
+  { "unpckhps", XM, EX },
+  { "movhps", XM, Ev },
+  { "movhps", Ev, XM },  
   /* 18 */
-  { "(bad)" },  { "(bad)" },  { "(bad)" },  { "(bad)" },  
+  { GRP14 },
+  { "(bad)" },  { "(bad)" },  { "(bad)" },  
   { "(bad)" },  { "(bad)" },  { "(bad)" },  { "(bad)" },  
   /* 20 */
   /* these are all backward in appendix A of the intel book */
@@ -582,11 +624,17 @@ static struct dis386 dis386_twobyte[] = {
   { "movl", Td, Rd },
   { "(bad)" },  
   /* 28 */
-  { "(bad)" },  { "(bad)" },  { "(bad)" },  { "(bad)" },  
-  { "(bad)" },  { "(bad)" },  { "(bad)" },  { "(bad)" },  
+  { "movaps", XM, EX },
+  { "movaps", EX, XM },
+  { PREGRP2 },
+  { "movntps", Ev, XM },
+  { PREGRP4 },  
+  { PREGRP3 },
+  { "ucomiss", XM, EX },
+  { "comiss", XM, EX },  
   /* 30 */
   { "wrmsr" },  { "rdtsc" },  { "rdmsr" },  { "rdpmc" },  
-  { "(bad)" },  { "(bad)" },  { "(bad)" },  { "(bad)" },  
+  { "sysenter" },  { "sysexit" },  { "(bad)" },  { "(bad)" },  
   /* 38 */
   { "(bad)" },  { "(bad)" },  { "(bad)" },  { "(bad)" },  
   { "(bad)" },  { "(bad)" },  { "(bad)" },  { "(bad)" },  
@@ -597,11 +645,23 @@ static struct dis386 dis386_twobyte[] = {
   { "cmovs", Gv,Ev }, { "cmovns", Gv,Ev }, { "cmovp", Gv,Ev }, { "cmovnp", Gv,Ev },
   { "cmovl", Gv,Ev }, { "cmovge", Gv,Ev }, { "cmovle", Gv,Ev }, { "cmovg", Gv,Ev },  
   /* 50 */
-  { "(bad)" },  { "(bad)" },  { "(bad)" },  { "(bad)" },  
-  { "(bad)" },  { "(bad)" },  { "(bad)" },  { "(bad)" },  
+  { "movmskps", Ev, XM },
+  { PREGRP13 },
+  { PREGRP12 },
+  { PREGRP11 },
+  { "andps", XM, EX },
+  { "andnps", XM, EX },
+  { "orps", XM, EX },
+  { "xorps", XM, EX },  
   /* 58 */
-  { "(bad)" },  { "(bad)" },  { "(bad)" },  { "(bad)" },  
-  { "(bad)" },  { "(bad)" },  { "(bad)" },  { "(bad)" },  
+  { PREGRP0 },
+  { PREGRP10 },
+  { "(bad)" },
+  { "(bad)" },  
+  { PREGRP14 },
+  { PREGRP7 },
+  { PREGRP5 },
+  { PREGRP6 },  
   /* 60 */
   { "punpcklbw", MX, EM },
   { "punpcklwd", MX, EM },
@@ -620,7 +680,7 @@ static struct dis386 dis386_twobyte[] = {
   { "movd", MX, Ev },
   { "movq", MX, EM },
   /* 70 */
-  { "(bad)" },
+  { "pshufw", MX, EM, Ib },
   { GRP10 },
   { GRP11 },
   { GRP12 },
@@ -685,7 +745,7 @@ static struct dis386 dis386_twobyte[] = {
   { "btsS", Ev, Gv },  
   { "shrdS", Ev, Gv, Ib },
   { "shrdS", Ev, Gv, CL },
-  { "(bad)" },
+  { GRP13 },
   { "imulS", Gv, Ev },  
   /* b0 */
   { "cmpxchgb", Eb, Gb },
@@ -708,11 +768,11 @@ static struct dis386 dis386_twobyte[] = {
   /* c0 */
   { "xaddb", Eb, Gb },
   { "xaddS", Ev, Gv },
-  { "(bad)" },
+  { PREGRP1 },
   { "(bad)" },  
-  { "(bad)" },
-  { "(bad)" },
-  { "(bad)" },
+  { "pinsrw", MX, Ev, Ib },
+  { "pextrw", Ev, MX, Ib },
+  { "shufps", XM, EX, Ib },
   { GRP9 },  
   /* c8 */
   { "bswap", eAX },
@@ -730,32 +790,34 @@ static struct dis386 dis386_twobyte[] = {
   { "psrlq", MX, EM },
   { "(bad)" },
   { "pmullw", MX, EM },
-  { "(bad)" },  { "(bad)" },  
+  { "(bad)" },
+  { "pmovmskb", Ev, MX },  
   /* d8 */
   { "psubusb", MX, EM },
   { "psubusw", MX, EM },
-  { "(bad)" },
+  { "pminub", MX, EM },
   { "pand", MX, EM },
   { "paddusb", MX, EM },
   { "paddusw", MX, EM },
-  { "(bad)" },
+  { "pmaxub", MX, EM },
   { "pandn", MX, EM },
   /* e0 */
-  { "(bad)" },
+  { "pavgb", MX, EM },
   { "psraw", MX, EM },
   { "psrad", MX, EM },
-  { "(bad)" },
-  { "(bad)" },
+  { "pavgw", MX, EM },
+  { "pmulhuw", MX, EM },
   { "pmulhw", MX, EM },
-  { "(bad)" },  { "(bad)" },  
+  { "(bad)" },
+  { "movntq", Ev, MX },  
   /* e8 */
   { "psubsb", MX, EM },
   { "psubsw", MX, EM },
-  { "(bad)" },
+  { "pminsw", MX, EM },
   { "por", MX, EM },
   { "paddsb", MX, EM },
   { "paddsw", MX, EM },
-  { "(bad)" },
+  { "pmaxsw", MX, EM },
   { "pxor", MX, EM },
   /* f0 */
   { "(bad)" },
@@ -764,7 +826,8 @@ static struct dis386 dis386_twobyte[] = {
   { "psllq", MX, EM },
   { "(bad)" },
   { "pmaddwd", MX, EM },
-  { "(bad)" },  { "(bad)" },  
+  { "psadbw", MX, EM },
+  { "maskmovq", MX, EM },  
   /* f8 */
   { "psubb", MX, EM },
   { "psubw", MX, EM },
@@ -796,22 +859,49 @@ static const unsigned char onebyte_has_modrm[256] = {
 };
 
 static const unsigned char twobyte_has_modrm[256] = {
-  /* 00 */ 1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0, /* 0f */
-  /* 10 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 1f */
-  /* 20 */ 1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0, /* 2f */
+  /*       0 1 2 3 4 5 6 7 8 9 a b c d e f        */
+  /*       -------------------------------        */
+  /* 00 */ 1,1,1,1,0,0,0,0,0,0,0,0,0,1,0,1, /* 0f */
+  /* 10 */ 1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0, /* 1f */
+  /* 20 */ 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, /* 2f */
   /* 30 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 3f */
   /* 40 */ 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, /* 4f */
-  /* 50 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 5f */
+  /* 50 */ 1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1, /* 5f */
   /* 60 */ 1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1, /* 6f */
-  /* 70 */ 0,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1, /* 7f */
+  /* 70 */ 1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1, /* 7f */
   /* 80 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 8f */
   /* 90 */ 1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1, /* 9f */
   /* a0 */ 0,0,0,1,1,1,1,1,0,0,0,1,1,1,1,1, /* af */
   /* b0 */ 1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1, /* bf */
   /* c0 */ 1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0, /* cf */
-  /* d0 */ 0,1,1,1,0,1,0,0,1,1,0,1,1,1,0,1, /* df */
-  /* e0 */ 0,1,1,0,0,1,0,0,1,1,0,1,1,1,0,1, /* ef */
-  /* f0 */ 0,1,1,1,0,1,0,0,1,1,1,0,1,1,1,0  /* ff */
+  /* d0 */ 0,1,1,1,0,1,0,1,1,1,1,1,1,1,1,1, /* df */
+  /* e0 */ 1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1, /* ef */
+  /* f0 */ 0,1,1,1,0,1,1,1,1,1,1,0,1,1,1,0  /* ff */
+  /*       -------------------------------        */
+  /*       0 1 2 3 4 5 6 7 8 9 a b c d e f        */
+};
+
+static const unsigned char twobyte_uses_f3_prefix[256] = {
+  /*       0 1 2 3 4 5 6 7 8 9 a b c d e f        */
+  /*       -------------------------------        */
+  /* 00 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 0f */
+  /* 10 */ 1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 1f */
+  /* 20 */ 0,0,0,0,0,0,0,0,0,0,1,0,1,1,0,0, /* 2f */
+  /* 30 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 3f */
+  /* 40 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 4f */
+  /* 50 */ 0,1,1,1,0,0,0,0,1,1,0,0,1,1,1,1, /* 5f */
+  /* 60 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 6f */
+  /* 70 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 7f */
+  /* 80 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 8f */
+  /* 90 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* 9f */
+  /* a0 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* af */
+  /* b0 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* bf */
+  /* c0 */ 0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0, /* cf */
+  /* d0 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* df */
+  /* e0 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* ef */
+  /* f0 */ 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0  /* ff */
+  /*       -------------------------------        */
+  /*       0 1 2 3 4 5 6 7 8 9 a b c d e f        */
 };
 
 static char obuf[100];
@@ -1061,6 +1151,117 @@ static struct dis386 grps[][8] = {
     { "(bad)" },
     { "psllq", MS, Ib },
     { "(bad)" },
+  },
+  /* GRP13 */
+  {
+    { "fxsave", Ev },
+    { "fxrstor", Ev },
+    { "ldmxcsr", Ev },
+    { "stmxcsr", Ev },
+    { "(bad)" },
+    { "(bad)" },
+    { "(bad)" },
+    { "sfence", None },
+  },
+  /* GRP14 */
+  {
+    { "prefetchnta", Ev },
+    { "prefetcht0", Ev },
+    { "prefetcht1", Ev },
+    { "prefetcht2", Ev },
+    { "(bad)" },
+    { "(bad)" },
+    { "(bad)" },
+    { "(bad)" },
+  }
+};
+
+static const char *simd_cmp_op [] = {
+  "eq",
+  "lt",
+  "le",
+  "unord",
+  "neq",
+  "nlt",
+  "nle",
+  "ord"
+};
+
+static struct dis386 prefix_user_table[][2] = {
+  /* PREGRP0 */
+  {
+    { "addps", XM, EX },
+    { "addss", XM, EX },
+  },
+  /* PREGRP1 */
+  {
+    { "cmp%sps", XM, EX, Ib },
+    { "cmp%sss", XM, EX, Ib },
+  },
+  /* PREGRP2 */
+  {
+    { "cvtpi2ps", XM, EM },
+    { "cvtsi2ss", XM, Ev },
+  },
+  /* PREGRP3 */
+  {
+    { "cvtps2pi", MX, EX },
+    { "cvtss2si", Gv, EX },
+  },
+  /* PREGRP4 */
+  {
+    { "cvttps2pi", MX, EX },  
+    { "cvttss2si", Gv, EX },  
+  },
+  /* PREGRP5 */
+  {
+    { "divps", XM, EX },
+    { "divss", XM, EX },
+  },
+  /* PREGRP6 */
+  {
+    { "maxps", XM, EX },
+    { "maxss", XM, EX },
+  },
+  /* PREGRP7 */
+  {
+    { "minps", XM, EX },
+    { "minss", XM, EX },
+  },
+  /* PREGRP8 */
+  {
+    { "movups", XM, EX },
+    { "movss", XM, EX },
+  },
+  /* PREGRP9 */
+  {
+    { "movups", EX, XM },
+    { "movss", EX, XM },
+  },
+  /* PREGRP10 */
+  {
+    { "mulps", XM, EX },
+    { "mulss", XM, EX },
+  },
+  /* PREGRP11 */
+  {
+    { "rcpps", XM, EX },
+    { "rcpss", XM, EX },
+  },
+  /* PREGRP12 */
+  {
+    { "rsqrtps", XM, EX },
+    { "rsqrtss", XM, EX },
+  },
+  /* PREGRP13 */
+  {
+    { "sqrtps", XM, EX },
+    { "sqrtss", XM, EX },
+  },
+  /* PREGRP14 */
+  {
+    { "subps", XM, EX },
+    { "subss", XM, EX },
   }
 };
 
@@ -1173,13 +1374,16 @@ print_insn_x86 (pc, info, aflag, dflag)
   char *first, *second, *third;
   int needcomma;
   unsigned char need_modrm;
+  unsigned char uses_f3_prefix;
+  unsigned char opcode;
+  int simd_cmp = 0;
 
   struct dis_private priv;
   bfd_byte *inbuf = priv.the_buffer;
 
   /* The output looks better if we put 5 bytes on a line, since that
      puts long word instructions on a single line.  */
-  info->bytes_per_line = 5;
+  info->bytes_per_line = 6;
 
   info->private_data = (PTR) &priv;
   priv.max_fetched = priv.the_buffer;
@@ -1209,9 +1413,7 @@ print_insn_x86 (pc, info, aflag, dflag)
     enter_instruction = 0;
   
   obufp = obuf;
-  
-  if (prefixes & PREFIX_REPZ)
-    oappend ("repz ");
+
   if (prefixes & PREFIX_REPNZ)
     oappend ("repnz ");
   if (prefixes & PREFIX_LOCK)
@@ -1242,13 +1444,19 @@ print_insn_x86 (pc, info, aflag, dflag)
       FETCH_DATA (info, codep + 2);
       dp = &dis386_twobyte[*++codep];
       need_modrm = twobyte_has_modrm[*codep];
+      uses_f3_prefix = twobyte_uses_f3_prefix[*codep];
     }
   else
     {
       dp = &dis386[*codep];
       need_modrm = onebyte_has_modrm[*codep];
+      uses_f3_prefix = 0;
     }
+  opcode = *codep;
   codep++;
+  
+  if (!uses_f3_prefix && (prefixes & PREFIX_REPZ))
+    oappend ("repz ");
 
   if (need_modrm)
     {
@@ -1265,7 +1473,20 @@ print_insn_x86 (pc, info, aflag, dflag)
   else
     {
       if (dp->name == NULL)
-	dp = &grps[dp->bytemode1][reg];
+        {
+	  switch(dp->bytemode2)
+	    {
+	      case USE_GROUPS:
+	        dp = &grps[dp->bytemode1][reg];
+		break;
+	      case USE_PREFIX_USER_TABLE:
+		simd_cmp = dp->bytemode1 == 1;
+		dp = &prefix_user_table[dp->bytemode1][prefixes & 0x01];
+		break;
+	      default:
+		return(1);
+	    }
+	}
       
       putop (dp->name, aflag, dflag);
       
@@ -1279,10 +1500,19 @@ print_insn_x86 (pc, info, aflag, dflag)
       if (dp->op2)
 	(*dp->op2)(dp->bytemode2, aflag, dflag);
       
-      obufp = op3out;
-      op_ad = 0;
-      if (dp->op3)
-	(*dp->op3)(dp->bytemode3, aflag, dflag);
+      if (simd_cmp)
+	{
+	  FETCH_DATA (the_info, codep + 1);
+	  sprintf (scratchbuf, obuf, simd_cmp_op [*codep++ & 0xff]);
+	  strcpy (obuf, scratchbuf);
+	}
+      else
+	{
+	  obufp = op3out;
+	  op_ad = 0;
+	  if (dp->op3)
+	    (*dp->op3)(dp->bytemode3, aflag, dflag);
+	}
     }
   
   obufp = obuf + strlen (obuf);
@@ -1713,6 +1943,20 @@ OP_indirE (bytemode, aflag, dflag)
   return OP_E (bytemode, aflag, dflag);
 }
 
+static int OP_PREFETCH(bytemode, aflag, dflag)
+     int bytemode;
+     int aflag;
+     int dflag;
+{
+  switch(reg) {
+    case 1:
+      obufp=obuf; oappend("prefetchw ");
+      break;
+    default:
+  }
+  return OP_E (bytemode, aflag, dflag);
+}
+	  
 static int
 OP_E (bytemode, aflag, dflag)
      int bytemode;
@@ -2097,7 +2341,7 @@ OP_DIR (size, aflag, dflag)
   switch (size) 
     {
     case lptr:
-      if (aflag) 
+      if (dflag) 
 	{
 	  offset = get32 ();
 	  seg = get16 ();
@@ -2110,8 +2354,9 @@ OP_DIR (size, aflag, dflag)
       sprintf (scratchbuf, "0x%x,0x%x", seg, offset);
       oappend (scratchbuf);
       break;
+#if 0
     case v_mode:
-      if (aflag)
+      if (dflag)
 	offset = get32 ();
       else
 	{
@@ -2125,6 +2370,7 @@ OP_DIR (size, aflag, dflag)
       sprintf (scratchbuf, "0x%x", offset);
       oappend (scratchbuf);
       break;
+#endif
     default:
       oappend ("<internal disassembler error>");
       break;
@@ -2272,6 +2518,17 @@ OP_MMX (bytemode, aflag, dflag)
   return 0;
 }
 
+static int
+OP_XMM (bytemode, aflag, dflag)
+     int bytemode;
+     int aflag;
+     int dflag;
+{
+  sprintf (scratchbuf, "%%xmm%d", reg);
+  oappend (scratchbuf);
+  return 0;
+}
+
 static int
 OP_EM (bytemode, aflag, dflag)
      int bytemode;
@@ -2287,6 +2544,21 @@ OP_EM (bytemode, aflag, dflag)
   return 0;
 }
 
+static int
+OP_EX (bytemode, aflag, dflag)
+     int bytemode;
+     int aflag;
+     int dflag;
+{
+  if (mod != 3)
+    return OP_E (bytemode, aflag, dflag);
+
+  codep++;
+  sprintf (scratchbuf, "%%xmm%d", rm);
+  oappend (scratchbuf);
+  return 0;
+}
+
 static int
 OP_MS (bytemode, aflag, dflag)
      int bytemode;
@@ -2298,3 +2570,110 @@ OP_MS (bytemode, aflag, dflag)
   oappend (scratchbuf);
   return 0;
 }
+
+static int
+OP_None (bytemode, aflag, dflag)
+     int bytemode;
+     int aflag;
+     int dflag;
+{
+  ++codep;
+  return 0;
+}
+
+static int
+OP_3DNow (bytemode, aflag, dflag)
+     int bytemode;
+     int aflag;
+     int dflag;
+{
+  sprintf (scratchbuf, "%%mm%d", reg);
+  oappend (scratchbuf);
+  return 0;
+}
+
+static char *_3DNowSuffixes[]= {
+/* 00 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 04 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 08 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 0C */	NULL,		 "pi2fd",	 NULL,		 NULL,		
+/* 10 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 14 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 18 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 1C */	NULL,		 "pf2id",	 NULL,		 NULL,		
+/* 20 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 24 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 28 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 2C */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 30 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 34 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 38 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 3C */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 40 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 44 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 48 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 4C */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 50 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 54 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 58 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 5C */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 60 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 64 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 68 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 6C */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 70 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 74 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 78 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 7C */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 80 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 84 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 88 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 8C */	NULL,		 NULL,		 NULL,		 NULL,		
+/* 90 */	"pfcmpge",	 NULL,		 NULL,		 NULL,		
+/* 94 */	"pfmin",	 NULL,		 "pfrcp",	 "pfrsqrt",	
+/* 98 */	NULL,		 NULL,		 "pfsub",	 NULL,		
+/* 9C */	NULL,		 NULL,		 "pfadd",	 NULL,		
+/* A0 */	"pfcmpgt",	 NULL,		 NULL,		 NULL,		
+/* A4 */	"pfmax",	 NULL,		 "pfrcpit1",	 "pfrsqit1",	
+/* A8 */	NULL,		 NULL,		 "pfsubr",	 NULL,		
+/* AC */	NULL,		 NULL,		 "pfacc",	 NULL,		
+/* B0 */	"pfcmpeq",	 NULL,		 NULL,		 NULL,		
+/* B4 */	"pfmul",	 NULL,		 "pfrcpit2",	 "pfmulhrw",	
+/* B8 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* BC */	NULL,		 NULL,		 NULL,		 "pavgusb",	
+/* C0 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* C4 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* C8 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* CC */	NULL,		 NULL,		 NULL,		 NULL,		
+/* D0 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* D4 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* D8 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* DC */	NULL,		 NULL,		 NULL,		 NULL,		
+/* E0 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* E4 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* E8 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* EC */	NULL,		 NULL,		 NULL,		 NULL,		
+/* F0 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* F4 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* F8 */	NULL,		 NULL,		 NULL,		 NULL,		
+/* FC */	NULL,		 NULL,		 NULL,		 NULL,		
+};
+
+static int
+OP_3DNowSuffix (bytemode, aflag, dflag)
+     int bytemode;
+     int aflag;
+     int dflag;
+{
+  FETCH_DATA (the_info, codep + 1);
+  if (_3DNowSuffixes[*codep]) {
+	  /* HACK */
+    obufp=obuf;
+    oappend(_3DNowSuffixes[*codep++]);
+    oappend(" ");
+  } else {
+    sprintf (scratchbuf, "0x%x", *codep++);
+    oappend (scratchbuf);
+  }
+  return 0;
+}
